""" Network Manager CLI (nmcli) functions """

from __future__ import annotations
import typing as T
import subprocess
import logging
import pandas
import io
from time import sleep

from .cmd import get_nmcli


def cli_config_check() -> bool:
    # %% check that NetworkManager CLI is available and WiFi is active

    try:
        ret = subprocess.check_output([get_nmcli(), "-t", "radio", "wifi"], text=True, timeout=2)
    except subprocess.CalledProcessError as err:
        logging.error(err)
        return False

    stdout = ret.strip().split(":")
    if "enabled" in stdout:
        return True

    if "disabled" in stdout:
        logging.error(
            """must enable WiFi, perhaps via:
nmcli radio wifi on"""
        )
        return False

    logging.error("could not determine WiFi state.")
    return False


def get_signal() -> str:
    cmd = [get_nmcli(), "-g", "SSID,BSSID,FREQ,SIGNAL", "device", "wifi"]
    # Debian stretch, Ubuntu 18.04
    # cmd = [EXE, "-t", "-f", "SSID,BSSID,FREQ,SIGNAL", "device", "wifi"]
    # ubuntu 16.04

    try:
        subprocess.check_call(cmd, timeout=1.0)
    except subprocess.CalledProcessError as err:
        raise ConnectionError(f"could not connect with NetworkManager for WiFi   {err}")

    sleep(0.5)  # nmcli errored for less than about 0.2 sec.
    # takes several seconds to update, so do it now.

    scan = [get_nmcli(), "device", "wifi", "rescan"]

    try:
        ret = subprocess.check_output(scan, timeout=1.0, text=True)
    except subprocess.CalledProcessError as err:
        logging.error(f"consider slowing scan cadence. {err}")

    return ret


def parse_signal(raw: str) -> list[dict[str, T.Any]]:
    dat = pandas.read_csv(
        io.StringIO(raw),
        sep=r"(?<!\\):",
        index_col=False,
        header=0,
        encoding="utf8",
        engine="python",
        dtype=str,
        usecols=[0, 1, 3],
        names=["ssid", "macAddress", "signalStrength"],
    )
    # %% optout
    dat = dat[~dat["ssid"].str.endswith("_nomap")]

    # %% cleanup
    dat["ssid"] = dat["ssid"].str.replace("nan", "")
    dat["macAddress"] = dat["macAddress"].str.replace(r"\\:", ":")

    return dat



""" Windows NetSH functions """

from __future__ import annotations
import typing as T
import subprocess
import logging
import io

from .cmd import get_netsh


def cli_config_check() -> bool:
    # %% check that NetSH EXE is available and WiFi is active
    cmd = [get_netsh(), "wlan", "show", "networks", "mode=bssid"]

    try:
        ret = subprocess.check_output(cmd, text=True, timeout=2)
    except subprocess.CalledProcessError as err:
        logging.error(err)
        return False

    for line in ret.split("\n"):
        if "networks currently visible" in line:
            return True
        if (
            "The wireless local area network interface is powered down and doesn't support the requested operation"
            in line
        ):
            logging.error("must enable WiFi, it appears to be turned off.")
            return False

    logging.error("could not determine WiFi state.")
    return False


def get_signal() -> str:
    """
    get signal strength using EXE

    returns dict of data parsed from EXE
    """

    cmd = [get_netsh(), "wlan", "show", "networks", "mode=bssid"]
    try:
        ret = subprocess.check_output(cmd, timeout=1.0, text=True)
    except subprocess.CalledProcessError as err:
        logging.error(f"consider slowing scan cadence.  {err}")

    return ret


def parse_signal(raw: str) -> list[dict[str, T.Any]]:
    dat: list[dict[str, str]] = []
    out = io.StringIO(raw)

    for line in out:
        d: dict[str, str] = {}
        if not line.startswith("SSID"):
            continue
        ssid = line.split(":", 1)[1].strip()
        # optout
        if ssid.endswith("_nomap"):
            continue

        # find BSSID MAC address
        for line in out:
            if not line[4:9] == "BSSID":
                continue
            d["macAddress"] = line.split(":", 1)[1].strip()
            for line in out:
                if not line[9:15] == "Signal":
                    continue
                try:
                    signal_percent = int(line.split(":", 1)[1].split("%", 1)[0])
                except ValueError:
                    logging.error(line)
                    break
                d["signalStrength"] = str(signal_percent_to_dbm(signal_percent))
                d["ssid"] = ssid
                dat.append(d)
                d = {}
                # need break at each for level
                break
            break

    return dat


def signal_percent_to_dbm(percent: int) -> int:
    """
    arbitrary conversion factor from Windows WiFi signal % to dBm
    assumes signal percents map to dBm like:

    * 100% is -30 dBm
    * 0% is -100 dBm

    Parameters
    ----------
    percent: int
        signal strength as percent 0..100

    Returns
    -------
    meas_dBm: int
        truncate to nearest integer because of uncertainties
    """

    REF = -100  # dBm
    assert 0 <= percent <= 100, "percent must be 0...100"

    return int(REF + percent * 7 / 10)



from __future__ import annotations

import os
import functools
import shutil


def running_as_root() -> bool:
    return os.getuid() == 0


@functools.cache
def get_exe(name: str, path: str | None = None) -> str:
    if not (exe := shutil.which(name, path=path)):
        raise FileNotFoundError(name)
    return exe


def get_airport() -> str:
    return get_exe(
        "airport",
        "/System/Library/PrivateFrameworks/Apple80211.framework/Versions/Current/Resources",
    )


def get_nmcli() -> str:
    return get_exe("nmcli")


def get_netsh() -> str:
    return get_exe("netsh")



import sys


match sys.platform:
    case "win32":
        from .netsh import cli_config_check, get_signal, parse_signal
    case "linux":
        from .netman import cli_config_check, get_signal, parse_signal
    case "darwin":
        from .airport import cli_config_check, get_signal, parse_signal
    case _:
        raise ImportError(f"MozLoc doesn't work with platform {sys.platform}")


__all__ = ["cli_config_check", "get_signal", "parse_signal"]



""" MacOS airport functions """

from __future__ import annotations
import typing as T
import logging
import subprocess
import re

from .cmd import get_airport, running_as_root


def cli_config_check() -> bool:
    # %% check that Airport is available and WiFi is active

    try:
        ret = subprocess.check_output([get_airport(), "--getinfo"], text=True, timeout=30)
    except subprocess.CalledProcessError as err:
        logging.error(err)
        return False

    stdout = ret.strip().lower()
    if len(stdout) == "airport: off":
        return False

    if "state: running" in stdout:
        return True

    logging.error("could not determine WiFi state.")
    return False


def get_signal() -> str:
    try:
        ret = subprocess.check_output([get_airport(), "--scan"], text=True, timeout=30)
    except subprocess.CalledProcessError as err:
        logging.error(f"consider slowing scan cadence. {err}")

    return ret


def parse_signal(raw: str) -> list[dict[str, T.Any]]:
    isroot = running_as_root()

    psudo = r"\s*([0-9a-zA-Z\s\-\.]+)\s+([0-9a-f]{2}(?::[0-9a-f]{2}){5})\s+(-\d{2,3})"
    # BSSID only present if sudo
    puser = r"\s*([0-9a-zA-Z\s\-\.]+)\s+(-\d{2,3})"
    # non-sudo has no BSSID

    p = psudo if isroot else puser
    isig = 3 if isroot else 2
    ibssid = 2
    pat = re.compile(p)
    dat: list[dict[str, str]] = []

    for line in raw.split("\n"):
        mat = pat.match(line)
        if mat:
            # Hidden SSID optout implicitly excluded by regex
            ssid = mat.group(1).strip()
            # optout
            if ssid.endswith("_nomap"):
                continue
            d = {"ssid": ssid, "signalStrength": mat.group(isig)}
            if isroot:
                d["macAddress"] = mat.group(ibssid)
            dat.append(d)

    return dat
